# Copyright (c) Stephen Brennan <stephen@brennan.io>
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Helper for opening a file as root and transmitting it via unix socket"""
import argparse
import array
import errno
import os
from pathlib import Path
import socket
import subprocess
import sys
import tempfile
from typing import Union


def recv_fd(sock: socket.socket) -> int:
    """Receive a file descripter on the socket"""
    # Borrows heavily from the Python docs for sock.recvmsg()
    fds = array.array("i")
    msg, ancdata, flags, addr = sock.recvmsg(4096, socket.CMSG_SPACE(fds.itemsize))
    if msg == b"success":
        level, typ, data = ancdata[0]
        assert level == socket.SOL_SOCKET
        assert typ == socket.SCM_RIGHTS
        data = data[: fds.itemsize]
        fds.frombytes(data)
        return fds[0]
    raise Exception(msg.decode())


def recv_error(sock: socket.socket) -> None:
    """Try to receive an error message from the sudo helper"""
    sock.setblocking(False)
    try:
        msg, ancdata, flags, addr = sock.recvmsg(4096)
    except OSError as e:
        if e.errno in (errno.EAGAIN, errno.EWOULDBLOCK):
            raise Exception("sudo helper never connected to socket") from None
        raise
    raise Exception(msg.decode())


def send_fd(sock: socket.socket, fd: int) -> None:
    """Send a file descripter on the socket"""
    # Borrows heavily from the Python docs for sock.sendmsg()
    fds = array.array("i", [fd])
    sock.sendmsg(
        [b"success"],
        [(socket.SOL_SOCKET, socket.SCM_RIGHTS, fds)],
    )


def open_via_sudo(
    path: Union[Path, str],
    flags: int = os.O_RDONLY,
    mode: int = 0o777,
) -> int:
    """Implements os.open() using sudo to get permissions"""
    # Currently does not support dir_fd argument
    path = str(path)
    with tempfile.TemporaryDirectory() as td:
        sockpath = Path(td) / "sock"
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        sock.bind(str(sockpath))

        try:
            proc = subprocess.run(
                [
                    "sudo",
                    sys.executable,
                    "-B",
                    __file__,
                    str(sockpath),
                    path,
                    str(flags),
                    str(mode),
                ],
                check=False,
            )
            if proc.returncode == 0:
                return recv_fd(sock)
            else:
                recv_error(sock)
                assert False  # No return from recv_error
        finally:
            # sudo tends to mess up the terminal, get it into a sane state
            subprocess.run(["stty", "sane"])


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "socket",
        type=Path,
        help="path to unix domain socket",
    )
    parser.add_argument(
        "file",
        type=Path,
        help="filename to open",
    )
    parser.add_argument(
        "flags",
        type=int,
        help="numeric flags",
    )
    parser.add_argument(
        "mode",
        type=int,
        help="numeric mode",
    )
    args = parser.parse_args()

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    sock.connect(str(args.socket))
    try:
        fd = os.open(args.file, args.flags, args.mode)
        send_fd(sock, fd)
    except Exception as e:
        sock.sendmsg([str(e).encode()])
        sys.exit(1)


if __name__ == "__main__":
    main()
